@@ -12,6 +12,7 @@ import copy
 import json
 
 from ruamel.yaml.comments import CommentedMap
+from urllib.request import urlopen
 
 import jsonschema
 import pkgutil
@@ -107,9 +108,7 @@ def do_load(filename):
         if filename.endswith('.json'):
             return json.load(f)
 
-        # ruamel C loader doesn't support 1.2, but 1.1 is good enough for us
-        tmp = f.read().replace('YAML 1.2', 'YAML 1.1')
-        return yaml.load(tmp)
+        return yaml.load(f.read())
 
 def load_schema(schema):
     for path in schema_user_paths:
@@ -151,15 +150,7 @@ def _is_string_schema(subschema):
     return False
 
 def _extract_single_schemas(subschema):
-    tmpsch = {}
-
-    for match in ['const', 'enum', 'pattern', 'minimum', 'maximum']:
-        if not match in subschema:
-            continue
-        tmpsch[match] = subschema[match]
-        del subschema[match]
-
-    return tmpsch
+    return { k: subschema.pop(k) for k in ('const', 'enum', 'pattern', 'minimum', 'maximum') if k in subschema }
 
 def _fixup_string_to_array(propname, subschema):
     # nothing to do if we don't have a set of string schema
@@ -210,6 +201,9 @@ def _fixup_int_array_min_max_to_matrix(propname, subschema):
                 subschema = item
                 break
 
+    if 'items' in subschema and isinstance(subschema['items'],list):
+        return
+
     if _is_matrix_schema(subschema):
         return
 
@@ -235,6 +229,7 @@ def _fixup_int_array_min_max_to_matrix(propname, subschema):
         _fixup_items_size(subschema['oneOf'])
 
 def _fixup_int_array_items_to_matrix(propname, subschema):
+    itemkeys = ('items','minItems','maxItems', 'uniqueItems', 'default')
     if not is_int_array_schema(propname, subschema):
         return
 
@@ -249,17 +244,10 @@ def _fixup_int_array_items_to_matrix(propname, subschema):
         return
 
     if isinstance(subschema['items'],dict):
-        subschema['items'] = copy.deepcopy(subschema)
-        # Don't copy 'allOf'
-        subschema['items'].pop('allOf', None)
-        subschema['items'].pop('oneOf', None)
-        for k in list(subschema.keys()):
-            if k in ['items', 'allOf', 'oneOf']:
-                continue
-            subschema.pop(k)
+        subschema['items'] = {k: subschema.pop(k) for k in itemkeys if k in subschema}
 
     if isinstance(subschema['items'],list):
-        subschema['items'] = [ {'items': subschema['items']} ]
+        subschema['items'] = [ {k: subschema.pop(k) for k in itemkeys if k in subschema} ]
 
 def _fixup_scalar_to_array(propname, subschema):
     if not _is_int_schema(subschema):
@@ -298,15 +286,13 @@ def fixup_vals(propname, schema):
     # Now we should be a the schema level to do actual fixups
 #    print(schema)
 
+    schema.pop('description', None)
+
     # This can be removed once draft 2019.09 is supported
-    if '$ref' in schema and \
-        ((len(schema) > 1 and not 'description' in schema) or \
-        (len(schema) > 2 and 'description' in schema)):
+    if '$ref' in schema and (len(schema) > 1):
         schema['allOf'] = [ {'$ref': schema['$ref']} ]
         schema.pop('$ref')
 
-    schema.pop('description', None)
-
     _fixup_int_array_min_max_to_matrix(propname, schema)
     _fixup_int_array_items_to_matrix(propname, schema)
     _fixup_string_to_array(propname, schema)
@@ -346,7 +332,7 @@ def fixup_sub_schema(schema, is_prop):
         fixup_node_props(schema)
 
     for k,v in schema.items():
-        if k in ['select', 'if', 'then', 'else', 'additionalProperties']:
+        if k in ['select', 'if', 'then', 'else', 'additionalProperties', 'not']:
             fixup_sub_schema(v, False)
 
         if k in ['allOf', 'anyOf', 'oneOf']:
@@ -364,8 +350,8 @@ def fixup_sub_schema(schema, is_prop):
 def fixup_node_props(schema):
     if not {'properties', 'patternProperties'} & schema.keys():
         return
-    if not ('additionalProperties' in schema and schema['additionalProperties'] is not True) and \
-       not ('unevaluatedProperties' in schema and schema['unevaluatedProperties'] is not True):
+    if ('additionalProperties' in schema and schema['additionalProperties'] is True) or \
+       ('unevaluatedProperties' in schema and schema['unevaluatedProperties'] is True):
         return
 
     schema.setdefault('properties', dict())
@@ -482,7 +468,7 @@ def add_select_schema(schema):
             if len(compatible_list) != 0:
                 schema['select'] = {
                     'required': ['compatible'],
-                    'properties': {'compatible': {'contains': {'enum': list(compatible_list)}}}}
+                    'properties': {'compatible': {'contains': {'enum': sorted(compatible_list)}}}}
 
                 return
 
@@ -574,6 +560,7 @@ def process_schema(filename):
     if not 'select' in schema:
         return
 
+    schema["type"] = "object"
     schema["$filename"] = filename
     return schema
 
@@ -642,7 +629,7 @@ def http_handler(uri):
                 return load_schema(uri.replace(schema_base_url, ''))
             return process_schema(uri.replace(schema_base_url, ''))
 
-        return yaml.load(jsonschema.compat.urlopen(uri).read().decode('utf-8'))
+        return yaml.load(urlopen(uri).read().decode('utf-8'))
     except FileNotFoundError as e:
         print('Unknown file referenced:', e, file=sys.stderr)
         exit(-1)
@@ -677,20 +664,65 @@ class DTValidator(DTVal):
     resolver = jsonschema.RefResolver('', None, handlers=handlers)
     format_checker = jsonschema.FormatChecker()
 
-    def __init__(self, schema, types=()):
-        jsonschema.Draft7Validator.__init__(self, schema, types, resolver=self.resolver,
-                                            format_checker=self.format_checker)
+    def __init__(self, *args, **kwargs):
+        kwargs.setdefault('resolver', self.resolver)
+        kwargs.setdefault('format_checker', self.format_checker)
+        super().__init__(*args, **kwargs)
+
+    @classmethod
+    def annotate_error(self, error, schema, path):
+        error.note = None
+        error.schema_file = None
+
+        for e in error.context:
+            self.annotate_error(e, schema, path + e.schema_path)
+
+        scope = self.ID_OF(schema)
+        self.resolver.push_scope(scope)
+        ref_depth = 1
+
+        lastp = ''
+        for p in path:
+            # json-schema 3.2.0 includes 'if' in schema path
+            if lastp != 'properties' and p == 'if':
+                continue
+            lastp = p
+
+            while '$ref' in schema and isinstance(schema['$ref'], str):
+                ref = self.resolver.resolve(schema['$ref'])
+                schema = ref[1]
+                self.resolver.push_scope(ref[0])
+                ref_depth += 1
+
+            if '$id' in schema and isinstance(schema['$id'], str):
+                error.schema_file = schema['$id']
+
+            schema = schema[p]
+
+            if isinstance(schema, dict):
+                if 'description' in schema and isinstance(schema['description'], str):
+                    error.note = schema['description']
+
+        while ref_depth > 0:
+            self.resolver.pop_scope()
+            ref_depth -= 1
+
+        if isinstance(error.schema, dict) and 'description' in error.schema:
+            error.note = error.schema['description']
 
     @classmethod
     def iter_schema_errors(cls, schema):
         meta_schema = cls.resolver.resolve_from_url(schema['$schema'])
         for error in cls(meta_schema).iter_errors(schema):
+            cls(meta_schema).annotate_error(error, meta_schema, error.schema_path)
             error.linecol = get_line_col(schema, error.path)
             yield error
 
     def iter_errors(self, instance, _schema=None):
-        for error in jsonschema.Draft7Validator.iter_errors(self, instance, _schema):
+        for error in super().iter_errors(instance, _schema):
             error.linecol = get_line_col(instance, error.path)
+            error.note = None
+            error.schema_file = None
             yield error
 
     @classmethod
@@ -793,6 +825,8 @@ def format_error(filename, error, prefix="", nodename=None, verbose=False):
                 msg += '\n' + format_error(filename, suberror, prefix=prefix+"\t", nodename=nodename, verbose=verbose)
             elif not suberror.message in msg:
                 msg += '\n' + prefix + '\t' + suberror.message
+                if suberror.note and suberror.note != error.note:
+                    msg += '\n\t\t' + prefix + 'hint: ' + suberror.note
 
     elif error.schema_path[-1] == 'oneOf':
         msg = 'More than one condition true in oneOf schema:\n\t' + \
@@ -801,4 +835,10 @@ def format_error(filename, error, prefix="", nodename=None, verbose=False):
     else:
         msg = error.message
 
+    if error.note:
+        msg += '\n\t' + prefix + 'hint: ' + error.note
+
+    if error.schema_file:
+        msg += '\n\t' + prefix + 'from schema $id: ' + error.schema_file
+
     return src + msg
